#include "vloc_measure.h"

namespace Localization {
namespace Ego {
double VLocMeasure::latestUpdateTime_ = -1.0;

VLocMeasure::VLocMeasure(const MeasurementType &mType, const RecLocInfo &vLocInfo)
:BaseMeasure(mType)
{
  status_ = vLocInfo.status_;
//  vIntegratedState_ = vLocInfo.integratedState;//暂时不用这个状态

  stamp_ = vLocInfo.stamp_;
  if (stamp_ - latestUpdateTime_ < 1e-5) {
      isVlocBlock_ = true;
      std::cout<<std::setprecision(15)<<"VLocMeasure():this frame is repeated,stamp"<<stamp_
               <<",latestUpdateTime_"<<latestUpdateTime_<<std::endl;
  }
  latestUpdateTime_ = stamp_;
  posLLA_ = vLocInfo.posLLA_;
  posUTM_ = vLocInfo.posUTM_;
  attitude_ = vLocInfo.attitude_;//todo:视觉的yaw角要减去真北和网北误差，然后才能给ego当作观测
  {
    vLocPos_ = posUTM_;
    vLocAtt_ = attitude_;
  }
  Eigen::Vector3d ypr(attitude_.z(), 0.0, 0.0);
  quatUTM_ = EulerYPR2RotYPR(ypr);

  mRm_ = Eigen::MatrixXd::Zero(6,6);
  mRm_.block<3,3>(0,0) = rNoiseVar_ * Eigen::Matrix3d::Identity();
  mRm_.block<3,3>(3,3) = pNoiseVar_ * Eigen::Matrix3d::Identity();
  score_ = vLocInfo.score_;
  CheckValid();
}
void VLocMeasure::CheckValid()
{
  isVlocValid_ = true;
  lastVlocStamp_ = stamp_;

  // status
  if (status_ == RecLocInfo::RecLocStatus::good) {
    isVlocValid_ = true;//当前只要good就使用
    std::cout<<std::setprecision(15)<<"VLocMeasure-CheckValid():vLocStatus is good,stamp:"<<stamp_<<
      ",角度:"<<attitude_.transpose() * RAD2DEG<<std::endl;
  } else {
    std::cout<<std::setprecision(15)<<"VLocMeasure-CheckValid():vLocStatus is not good,isVlocValid_ false,stamp:"<<
      stamp_<<std::endl;
    isVlocValid_ = false; // compute status
  }


//  // pos jump//todo:改成ecef坐标系之后，用车体下的横向和纵向来检查
//  static Eigen::Vector3d lastPos = quatUTM_.inverse() * posUTM_;
//  Eigen::Vector3d curPos = quatUTM_.inverse() * posUTM_;
//  double dx = fabs(curPos.x() - lastPos.x());
//  double dy = fabs(curPos.y() - lastPos.y());
//  if (dx > 2.5 || dy > 1) { // 2.5m/frame, 180km/h
//    std::cerr<<std::setprecision(15)<<"VLocMeasure-CheckValid():Vloc pos jumps"
//      ",diff x:"<<dx<<",diff y:"<<dy<<",isVlocValid_ false"<<std::endl;
//    std::cout<<std::setprecision(15)<<"VLocMeasure-CheckValid():origin Vloc pos:"
//      <<posUTM_<<",att:"<<quatUTM_.toRotationMatrix().eulerAngles(0,1,2)*RAD2DEG<<std::endl;
//    isVlocValid_ = false;
//  }
//  lastPos = curPos;
//  // pos jump 3d norm,后续删掉
//  static Eigen::Vector3d lastPos = posUTM_;
//  Eigen::Vector3d curPos = posUTM_;
//  double diff = (curPos - lastPos).norm();
//  if (diff > 5) { // 2.5m/frame, 180km/h
//    std::cerr<<std::setprecision(15)<<"VLocMeasure-CheckValid():Vloc pos jumps, diff is:"<<
//      diff<<",isVlocValid_ false"<<std::endl;
//    isVlocValid_ = false;
//  }
//  lastPos = curPos;
  // yaw jump //todo:讲道理位置跳变和角度跳变要打开，目前关上只是因为vloc结果不那么离谱而已
//  static double lastYaw = attitude_.z();
//  double curYaw = attitude_.z();
//  if (fabs(curYaw - lastYaw) > 0.1) { // 5.72degree/frame, 114/s
//    std::cerr<<std::setprecision(15)<<"VLocMeasure-CheckValid():Vloc Yaw jumps,isVlocValid_ false"<<std::endl;
//    isVlocValid_ = false;
//  }
//  lastYaw = curYaw;
}

bool VLocMeasure::Update(const State &state, State &nextState)
{
  nextState = state;
  // return false;
  if (isVlocValid_ == false) {
    std::cout<<"VLoc-Update():isVlocValid_ false,can't update,use next pridict"<<std::endl;
    return false;
  }
  nextState.time = stamp_;

  Eigen::VectorXd innov(6);
  Eigen::MatrixXd mH = Eigen::MatrixXd::Zero(6,StateIndex::STATE_TOTAL);
  ObserveVloc(state,nextState,innov,mH);
  UpdateStateVLoc(state,nextState,innov,mH);
  return true;
}

void VLocMeasure::ObserveVloc(const State &state, State &nextState, Eigen::VectorXd &innov,
                                     Eigen::MatrixXd &mH)
{
  innov.head(3) = 2.0 * (quatUTM_.inverse() * state.q).vec();
  innov.tail(3) = posUTM_ - state.p;
  mH.block<3, 3>(0, StateIndex::R) =
          -1.0 * Qleft(quatUTM_.inverse() * state.q).bottomRightCorner<3, 3>();
  mH.block<3, 3>(3, StateIndex::P) = Eigen::Matrix3d::Identity();
}

bool VLocMeasure::UpdateStateVLoc(const State &state, State &nextState, const Eigen::VectorXd &innov,
                                 const Eigen::MatrixXd &mH)
{
  Eigen::MatrixXd mS = mH * state.mP * mH.transpose() + mRm_;
  Eigen::MatrixXd mSInv = mS.inverse();
  Eigen::MatrixXd mK = state.mP * mH.transpose() * mSInv;
  Eigen::VectorXd dx = mK * innov;
  Eigen::MatrixXd mSS = mH * state.mP * mH.transpose() + mRm_ * 10.0 * 10.0;

  UpdateState(state, nextState, dx, mH, mK);
  return true;
}

Eigen::VectorXd VLocMeasure::MeasurementEquation(const State &state) const
{
  Eigen::VectorXd measurement(6);
  measurement.head(3) = 2.0 * (quatUTM_.inverse() * state.q).vec();
  measurement.tail(3) = posUTM_ - state.p;
  return -measurement;
}

} // namespace Ego
} // namespace Localizaiton

